
///////////////////////////////////////////////////////////////////////////////
// File       : List.hpp
//
// Description: A Linked List template for C++.
//      This file has been modified
//      away from the ANSI standard template declarations to the
//      previous ANSI draft standard that the AIX xlC compiler
//      supports.
//
// Author     : James Michael Matthew Hare
//
// Last Mod   : March 20th, 2000
//
// Version    : 1.5
//
// History    : 1.0 - Initial version for ANSI standard C++ templates.
//      Designed to compile under Microsoft Visual C++ 6.0 SP3
//
///////////////////////////////////////////////////////////////////////////JMMH


// ifndef block to keep file from being included twice...
#ifndef LIST_HEADER_INCLUDED
#define LIST_HEADER_INCLUDED

// need assert for cin, cout
#include <iostream.h>



// Exception class for out of memory
class OutOfMemory
{
public:
    const char * what() const 
    { 
        return "Not enough memory to satisfy allocation request."; 
    }
};



// Exception class for bad iterator position
class InvalidIterator
{
public:
    const char * what() const 
    { 
        return "Iterator is not pointing to valid location in the list."; 
    }
};



// Exception class for illegal iterator for this list
class IllegalIterator
{
public:
    const char * what() const 
    { 
        return "Iterator provided does not operate over list it was passed to.";
    }
};



// Our linked list class.  Template, so can hold any type we desire
template <typename T>
class List
{
	// I'm putting Node in here, this way, only the class can see it.
    struct Node
    {
        T data;
        Node *next;
        Node *prev;
    };

    // Want a head, current, and previous.
    Node *head;         // Front of list
    Node *tail;         // Rear of list

	// generic insert function
	void genericInsert(Node *nodeBefore, Node *nodeAfter, const T& insertMe);

public:
    // make sure the iterator is our friend
    class  Iterator;
    friend Iterator;
    
    // the actual iterator.
	class Iterator
    {
        Node   *current;        // ptr to current node in linked list
        const List<T> *theList;

        // state list as our friend
        friend List<T>;

        // private constructor to create an iterator over a list and 
        // pt to some node
        Iterator(const List<T> *ls, Node *currNode)
		{ 
            theList = ls; 
            current = currNode; 
        }

    public:
        // constructors.  Default constr creates an invalid
        // ptr(pts to nothing).
        Iterator(void)
        { 
            theList = NULL;  
            current = NULL; 
        }

        // get methods to retrieve data at current location
        T &getData(void)
        { 
            if(!isValid())
                throw InvalidIterator(); 
            return current->data; 
        }

        const T &getData(void) const
        { 
            if(!isValid())
                throw InvalidIterator();
            return current->data; 
        }

        // true if pointing to a valid item
        bool isValid(void) const
            { return (current != NULL); }

        bool isOverList(List<T> &ls) const
            { return (&ls == theList); }

        // move the current item back and forth.
        void incr(void)
        { 
            if(!isValid())
                throw InvalidIterator(); 
            current = current->next; 
        }

        void decr(void)
        { 
            if(!isValid())
                throw InvalidIterator(); 
            current = current->prev; 
        }

        // operator overloads
        T& operator *(void)
        { 
            return getData(); 
        }

        const T& operator *(void) const
        { 
            return getData(); 
        }

        Iterator & operator++(void)
            { incr(); return *this; }

        Iterator & operator--(void)
            { decr(); return *this; }
    }; 

    // default constructor
    List(void);

    // Need copy constructor, explain why with parameter passing...
    List(const List<T> &oldList);

    // Need assignment operator.  Explain why with assignment...
    List<T> & operator=(const List<T> &oldList);

    // methods to get an iterator
    Iterator begin(void) const { return Iterator(this, head); }
    Iterator end(void) const   { return Iterator(this, tail); }

    // method to insert at different parts of the list.
    // All insertion methods return an iterator to the new item.
    void insertFront(const T &itemToInsert);
    void insertRear(const T &itemToInsert);
    void insertBefore(Iterator &it, const T& insertMe);
    void insertAfter(Iterator &it, const T& insertMe);

    // methods to delete at different parts of list.
    void removeFront(void);         // returns iterator to new head
    void removeRear(void);          // returns iterator to new tail
    void remove(Iterator &it);  // returns next item in line

    // return the size
    int size(void) const;

    // copy copies one linked list into another...
    void copy(const List<T> &oldList);

    // clear removes all items in the list...
    void clear(void);
    bool empty(void) const;

    // destroy list...
    virtual ~List(void);
};





// Default constructor... Create empty list...
template <typename T>
List<T> :: List(void)
{
    head = NULL;
    tail = NULL;
}




// check if list is empty
template <typename T>
bool List<T> :: empty(void) const
{
    return (head == NULL);
}




// Copy constructor...
template <typename T>
List<T> :: List(const List<T> &old)
: head(NULL),
  tail(NULL)
{
    // copy the old list, this is a function down below.
    copy(old);
}



// assignment... Should do same as copy constructor...  returns current object.
template <typename T>
List<T> & List<T> :: operator=(const List<T> &old)
{
	// copy the old list
    if(this != &old)
        copy(old);

    return *this;
}




// get size of list
template <typename T>
int List<T> :: size(void) const
{
	// start at zero.
    int Size = 0;

	// walk iterator through list
	Iterator countIt = begin();

	// as long as the it is valid, bump the size by one
	for(; countIt.isValid(); ++countIt)
		Size++;
    
    // return the count
    return Size;
}






// insert and delete manipulators to alter the list
template <typename T>
void List<T> :: remove(Iterator &it)
{
    // make sure our iterator is valid
    if(!it.isValid())
        throw InvalidIterator();
    if(!it.isOverList(*this))
        throw IllegalIterator();

    // set up three ptrs for delete
    Node *before   = it.current->prev;
    Node *after    = it.current->next;

    // reroute ptr to object before.  If no object before, reroute head
    if(before)
        before->next = after;
    else
        head = after;

    // reroute ptr from object after.  If no object after, do nothing
    if(after)
        after->prev = before;
    else
        tail = before;

    // now can kill current.
    delete it.current;

    // set current at after position and continue.
    it.current = after;

    // return anonymous iterator to current node
    it;
}



// generic insertion algorithm
template <typename T>
void List<T> :: genericInsert(Node *nodeBefore, Node *nodeAfter, const T& insertMe)
{
    // create new node and init fields.
    Node *current = new Node;

    if(!current)
        throw OutOfMemory();

    current->data = insertMe;
    current->prev = nodeBefore;
    current->next = nodeAfter;

    // set previous node to point to us next, or head if no previous node
    if(nodeBefore)
        nodeBefore->next = current;
    else
        head = current;

    if(nodeAfter)
        nodeAfter->prev = current;
    else
        tail = current;
}




// insert item before current iterator
template <typename T>
void List<T> :: insertBefore(Iterator &it, const T& insertMe)
{
    // make sure our iterator is valid
    if(!it.isValid())
        throw InvalidIterator();
    if(!it.isOverList(*this))
        throw IllegalIterator();

    // the node before is the one previous to current
    // once we insert a new node, it will be current and after will be
    // the next node (the old current);
	genericInsert(it.current->prev, it.current, insertMe);
}





template <typename T>
void List<T> :: insertAfter(Iterator &it, const T& insertMe)
{
    // make sure our iterator is valid
    if(!it.isValid())
        throw InvalidIterator();
    if(!it.isOverList(*this))
        throw IllegalIterator();

    // if we are inserting after the current node, then curent becomes
    // the before, after becomes the next.
    genericInsert(it.current, it.current->next, insertMe);
}



// method to delete at front of list
template <typename T>
void List<T>::removeFront(void)
{
    // delete the item indicated, which is the front item in the list
    remove(begin());
}



// method to delete at rear of list
template <typename T>
void List<T>::removeRear(void)
{
    // delete the item indicated, which is the rear item in the list
    remove(end());
}



// method to insert at front
template <typename T>
void List<T>::insertFront(const T &itemToInsert)
{
    // do generic insert on head
    genericInsert(NULL, head, itemToInsert);
}



// method to insert at rear of list
template <typename T>
void List<T>::insertRear(const T &itemToInsert)
{
    // do generic insert on tail
    genericInsert(tail, NULL, itemToInsert);
}



// copy one list to another
template <typename T>
void List<T> :: copy(const List<T> &old)
{
    // clear current list
    clear();

    // point to start of old list
    Iterator pos = old.begin();

    // copy all items over
    while(pos.isValid())
    {
        insertRear(pos.getData());
        ++pos;
    }
}



// Destroy contents of list
template <typename T>
void List<T> :: clear(void)
{
    // delete the list node by node.
    while(!empty())
        removeFront();
}




// delete the entire list before destruction
template <typename T>
List<T> :: ~List(void)
{
    // simply empty the list.
    clear();
}

#endif
